-------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, unb1_board_lib, correlator_lib, diag_lib, dp_lib, rTwoSDF_lib, wpfb_lib, st_lib, filter_lib, fft_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE unb1_board_lib.unb1_board_pkg.ALL;
USE diag_lib.diag_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE rTwoSDF_lib.rTwoSDFPkg.ALL; 
USE st_lib.ALL; 
USE filter_lib.ALL;
USE filter_lib.fil_pkg.ALL;
USE fft_lib.ALL;
USE fft_lib.fft_pkg.ALL; 
USE wpfb_lib.wpfb_pkg.ALL;

ENTITY apertif_unb1_correlator IS
  GENERIC (
    g_design_name : STRING  := "apertif_unb1_correlator";
    g_sim         : BOOLEAN := FALSE; --Overridden by TB
    g_sim_unb_nr  : NATURAL := 0;
    g_sim_node_nr : NATURAL := 0;
    g_stamp_date  : NATURAL := 0;  -- Date (YYYYMMDD) -- set by QSF
    g_stamp_time  : NATURAL := 0;  -- Time (HHMMSS)   -- set by QSF
    g_stamp_svn   : NATURAL := 0   -- SVN revision    -- set by QSF
  );
  PORT (
    -- GENERAL
    CLK          : IN    STD_LOGIC; -- System Clock
    PPS          : IN    STD_LOGIC; -- System Sync
    WDI          : OUT   STD_LOGIC; -- Watchdog Clear
    INTA         : INOUT STD_LOGIC; -- FPGA interconnect line
    INTB         : INOUT STD_LOGIC; -- FPGA interconnect line

    -- Others
    VERSION      : IN    STD_LOGIC_VECTOR(c_unb1_board_aux.version_w-1 DOWNTO 0);
    ID           : IN    STD_LOGIC_VECTOR(c_unb1_board_aux.id_w-1 DOWNTO 0);
    TESTIO       : INOUT STD_LOGIC_VECTOR(c_unb1_board_aux.testio_w-1 DOWNTO 0);
    
    -- I2C Interface to Sensors
    sens_sc      : INOUT STD_LOGIC;
    sens_sd      : INOUT STD_LOGIC;
  
    -- 1GbE Control Interface
    ETH_clk      : IN    STD_LOGIC;
    ETH_SGIN     : IN    STD_LOGIC;
    ETH_SGOUT    : OUT   STD_LOGIC
  );
END apertif_unb1_correlator;


ARCHITECTURE str OF apertif_unb1_correlator IS

  -- Firmware version x.y
  CONSTANT c_fw_version             : t_unb1_board_fw_version := (0, 1);
  -- In simulation we don't need the 1GbE core for MM control, deselect it in c_use_phy based on g_sim
  CONSTANT c_use_phy                : t_c_unb1_board_use_phy  := (sel_a_b(g_sim, 0, 1), 0, 0, 0, 0, 0, 0, 1);

  -- System
  SIGNAL cs_sim                     : STD_LOGIC;
  SIGNAL xo_clk                     : STD_LOGIC;
  SIGNAL xo_rst                     : STD_LOGIC;
  SIGNAL xo_rst_n                   : STD_LOGIC;
  SIGNAL mm_clk                     : STD_LOGIC;
  SIGNAL mm_locked                  : STD_LOGIC;
  SIGNAL mm_rst                     : STD_LOGIC;
  
  SIGNAL dp_rst                     : STD_LOGIC;
  SIGNAL dp_clk                     : STD_LOGIC;
  
  -- PIOs
  SIGNAL pout_wdi                   : STD_LOGIC;

  -- WDI override
  SIGNAL reg_wdi_mosi               : t_mem_mosi;
  SIGNAL reg_wdi_miso               : t_mem_miso;

  -- PPSH
  SIGNAL reg_ppsh_mosi              : t_mem_mosi;
  SIGNAL reg_ppsh_miso              : t_mem_miso;
  
  -- UniBoard system info
  SIGNAL reg_unb_system_info_mosi   : t_mem_mosi;
  SIGNAL reg_unb_system_info_miso   : t_mem_miso;
  SIGNAL rom_unb_system_info_mosi   : t_mem_mosi;
  SIGNAL rom_unb_system_info_miso   : t_mem_miso;

  -- UniBoard I2C sens
  SIGNAL reg_unb_sens_mosi          : t_mem_mosi;
  SIGNAL reg_unb_sens_miso          : t_mem_miso;

  -- eth1g
  SIGNAL eth1g_tse_clk              : STD_LOGIC;
  SIGNAL eth1g_mm_rst               : STD_LOGIC;
  SIGNAL eth1g_tse_mosi             : t_mem_mosi;  -- ETH TSE MAC registers
  SIGNAL eth1g_tse_miso             : t_mem_miso;
  SIGNAL eth1g_reg_mosi             : t_mem_mosi;  -- ETH control and status registers
  SIGNAL eth1g_reg_miso             : t_mem_miso;
  SIGNAL eth1g_reg_interrupt        : STD_LOGIC;   -- Interrupt
  SIGNAL eth1g_ram_mosi             : t_mem_mosi;  -- ETH rx frame and tx frame memory
  SIGNAL eth1g_ram_miso             : t_mem_miso;
  
  -- Correlator
  CONSTANT c_nof_inputs             : NATURAL := 12; --2;
  CONSTANT c_nof_input_folds        : NATURAL := 1;
  CONSTANT c_nof_input_streams      : NATURAL := c_nof_inputs / pow2(c_nof_input_folds);
  CONSTANT c_nof_pre_mult_folds     : NATURAL := 1; 
  CONSTANT c_complex_data_w         : NATURAL := 8; -- !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!1
  CONSTANT c_conjugate              : BOOLEAN := TRUE;
  CONSTANT c_nof_channels           : NATURAL := 64;
  CONSTANT c_integration_period     : NATURAL := sel_a_b(g_sim, 0, 12208);

  CONSTANT c_nof_visibilities       : NATURAL := (c_nof_inputs*(c_nof_inputs+1))/2;

  -- WPFB
  CONSTANT c_wpfb_wb_factor         : NATURAL := 1;                    -- = default 1, wideband factor
  CONSTANT c_wpfb_nof_wb_streams    : NATURAL := c_nof_input_streams;  -- = 1, the number of parallel wideband streams. The fi
  CONSTANT c_wpfb_nof_chan          : NATURAL := 1;                    -- = default 0, defines the number of channels (=time-m
  CONSTANT c_wpfb_nof_points        : NATURAL := 64;                   -- = 1024, N point FFT
  CONSTANT c_wpfb_nof_taps          : NATURAL := 8;                    -- = 8 nof taps n the filter
  CONSTANT c_wpfb_in_dat_w          : NATURAL := 8; --6;                    -- = 8, number of input bits                           
  CONSTANT c_wpfb_out_dat_w         : NATURAL := 14; --12;                   -- = 14, number of output bits: in_dat_w + natural((cei
  CONSTANT c_wpfb_use_separate      : BOOLEAN := FALSE;                -- = false for complex input, true for two real inputs
  CONSTANT c_wpfb_use_reorder       : BOOLEAN := FALSE;

  CONSTANT c_wpfb : t_wpfb  := (c_wpfb_wb_factor, c_wpfb_nof_points, c_wpfb_nof_chan, c_wpfb_nof_wb_streams,
                                c_wpfb_nof_taps, c_wpfb_in_dat_w, 16, 16,
                                c_wpfb_use_reorder, c_wpfb_use_separate, 16, c_wpfb_out_dat_w, 18, 2, true, 56, 2,                      
                                c_fft_pipeline, c_fft_pipeline, c_fil_ppf_pipeline);

  CONSTANT c_wpfb_coefs_file_prefix : STRING  := "../../../../../UniBoard/trunk/Firmware/dsp/filter/build/data/coefs_wide1_p64_t8";

  -- Block generator
  CONSTANT c_bg_block_size          : NATURAL := c_nof_channels*pow2(c_nof_input_folds);
  CONSTANT c_bg_gapsize             : NATURAL := 0;

  -- Indicate the integration period with the sync. In the correlator, the
  -- integration period is forced to a minimum of c_nof_visibilities to
  -- allow folding the output onto one stream.
  -- . The sync pulse is only there for the human eye (wave window) -
  --   it is not used by the correlator.
  CONSTANT c_bg_blocks_per_sync     : NATURAL := largest(c_integration_period, c_nof_visibilities);

  CONSTANT c_bg_ctrl                : t_diag_block_gen := ('1',                      -- enable             
                                                           '0',                      -- enable_sync        
                                                          TO_UVEC(     c_bg_block_size, c_diag_bg_samples_per_packet_w),
                                                          TO_UVEC(c_bg_blocks_per_sync, c_diag_bg_blocks_per_sync_w),
                                                          TO_UVEC(        c_bg_gapsize, c_diag_bg_gapsize_w),
                                                          TO_UVEC(                   0, c_diag_bg_mem_low_adrs_w),
                                                          TO_UVEC(   c_bg_block_size-1, c_diag_bg_mem_high_adrs_w),
                                                          TO_UVEC(                   0, c_diag_bg_bsn_init_w));

  SIGNAL wpfb_snk_in_arr            : t_dp_sosi_arr(c_nof_input_streams-1 DOWNTO 0);
  SIGNAL wpfb_src_out_arr           : t_dp_sosi_arr(c_nof_input_streams-1 DOWNTO 0);
  SIGNAL correlator_src_out_arr     : t_dp_sosi_arr(1-1 DOWNTO 0);

  SIGNAL ram_fil_coefs_mosi         : t_mem_mosi;
  SIGNAL ram_fil_coefs_miso         : t_mem_miso;

  SIGNAL ram_diag_data_buf_mosi     : t_mem_mosi;
  SIGNAL ram_diag_data_buf_miso     : t_mem_miso;
  SIGNAL reg_diag_data_buf_mosi     : t_mem_mosi;
  SIGNAL reg_diag_data_buf_miso     : t_mem_miso; 

BEGIN

  -----------------------------------------------------------------------------
  -- Block generators
  -----------------------------------------------------------------------------
  u_mms_diag_block_gen : ENTITY diag_lib.mms_diag_block_gen
  GENERIC MAP (
    g_nof_output_streams => c_nof_input_streams,
    g_buf_dat_w          => 2*c_complex_data_w,
    g_buf_addr_w         => ceil_log2(TO_UINT(c_bg_ctrl.samples_per_packet)),
    g_file_name_prefix   => "../../../applications/apertif/apertif_unb1_correlator/src/hex/composite_signals",
    g_diag_block_gen_rst => c_bg_ctrl
  )
  PORT MAP (
    mm_rst           => mm_rst,
    mm_clk           => mm_clk,

    dp_rst           => dp_rst,
    dp_clk           => dp_clk,

    out_sosi_arr     => wpfb_snk_in_arr
  );

  -----------------------------------------------------------------------------
  -- To replace block generators: 10GbE receivers 0..2
  -----------------------------------------------------------------------------
  -- Placeholder

  -----------------------------------------------------------------------------
  -- To replace block generators: BSN aligner
  -- . Produces 3 aligned streams
  -----------------------------------------------------------------------------
  -- Placeholder

  -----------------------------------------------------------------------------
  -- To replace block generators: 
  -- . Extracts 
  -----------------------------------------------------------------------------
  -- Placeholder




  -----------------------------------------------------------------------------
  -- WPFB
  -----------------------------------------------------------------------------
  u_wpfb_unit : ENTITY wpfb_lib.wpfb_unit
  GENERIC MAP(
    g_wpfb              => c_wpfb,     
    g_stats_ena         => FALSE,
    g_use_bg            => FALSE,    
    g_file_index_arr    => array_init(0, 128, 1),    
    g_coefs_file_prefix => c_wpfb_coefs_file_prefix 
  )
  PORT MAP(
    dp_rst             => dp_rst,
    dp_clk             => dp_clk,
    mm_rst             => mm_rst,
    mm_clk             => mm_clk,  
    ram_fil_coefs_mosi => ram_fil_coefs_mosi,
    ram_fil_coefs_miso => ram_fil_coefs_miso,
    ram_st_sst_mosi    => c_mem_mosi_rst,
    ram_st_sst_miso    => OPEN, 
    reg_bg_ctrl_mosi   => c_mem_mosi_rst,
    reg_bg_ctrl_miso   => OPEN,
    ram_bg_data_mosi   => c_mem_mosi_rst,
    ram_bg_data_miso   => OPEN,
    in_sosi_arr        => wpfb_snk_in_arr,     
    out_sosi_arr       => wpfb_src_out_arr
  ); 

  -----------------------------------------------------------------------------
  -- Stream recorder to record the WPFB output stream to a file
  -----------------------------------------------------------------------------
  u_dp_stream_rec_play_wpfb_unit : ENTITY dp_lib.dp_stream_rec_play
  GENERIC MAP (
    g_sim            => TRUE,
    g_pass_through   => FALSE,
    g_rec_not_play   => TRUE,
    g_rec_play_file  => "../../../applications/apertif/apertif_unb1_correlator/tb/rec/wpfb_src_out_arr0.rec",
    g_record_invalid => FALSE
  )
  PORT MAP (
    dp_clk  => dp_clk,
    snk_in  => wpfb_src_out_arr(0),
    snk_out => OPEN,
    src_out => OPEN,
    src_in  => c_dp_siso_rdy
  );

  -----------------------------------------------------------------------------
  -- Correlator
  -----------------------------------------------------------------------------
  u_correlator : ENTITY correlator_lib.correlator
  GENERIC MAP (
    g_nof_input_streams  => c_nof_input_streams,
    g_input_unfold_factor => c_nof_input_folds,
    g_nof_pre_mult_folds => c_nof_pre_mult_folds,
    g_data_w             => 14, --c_complex_data_w,
    g_conjugate          => c_conjugate,
    g_nof_channels       => c_nof_channels,
    g_integration_period => c_integration_period
  )
  PORT MAP (
    clk         => dp_clk,
    rst         => dp_rst,

    snk_in_arr  => wpfb_src_out_arr,
    src_out_arr => correlator_src_out_arr
  );

  -----------------------------------------------------------------------------
  -- Stream recorder to record the correlator output stream to a file
  -- . The data buffer can only take snapshots.
  -----------------------------------------------------------------------------
  u_dp_stream_rec_play_correlator : ENTITY dp_lib.dp_stream_rec_play
  GENERIC MAP (
    g_sim            => TRUE,
    g_pass_through   => FALSE,
    g_rec_not_play   => TRUE,
    g_rec_play_file  => "../../../applications/apertif/apertif_unb1_correlator/tb/rec/correlator_src_out_arr0.rec",
    g_record_invalid => FALSE
  )
  PORT MAP (
    dp_clk  => dp_clk,
    snk_in  => correlator_src_out_arr(0),
    snk_out => OPEN,
    src_out => OPEN,
    src_in  => c_dp_siso_rdy
  );


--  -----------------------------------------------------------------------------
--  -- Data buffer to be read out by Python
--  -----------------------------------------------------------------------------
--  u_diag_data_buffer : ENTITY diag_lib.mms_diag_data_buffer
--  GENERIC MAP (    
--    g_nof_streams  => 1,
--    g_data_w       => 64,
--    g_data_type    => e_complex,
--    g_buf_nof_data => c_nof_visibilities,
--    g_buf_use_sync => TRUE 
--  )
--  PORT MAP (
--    mm_rst            => mm_rst,
--    mm_clk            => mm_clk,
--    dp_rst            => dp_rst,
--    dp_clk            => dp_clk,
--
--    ram_data_buf_mosi => ram_diag_data_buf_mosi,
--    ram_data_buf_miso => ram_diag_data_buf_miso,
--    reg_data_buf_mosi => reg_diag_data_buf_mosi,
--    reg_data_buf_miso => reg_diag_data_buf_miso,
--
--    in_sync           => correlator_src_out_arr(0).sop,
--    in_sosi_arr       => correlator_src_out_arr
--  );

  -----------------------------------------------------------------------------
  -- General control function
  -----------------------------------------------------------------------------
  u_ctrl : ENTITY unb1_board_lib.ctrl_unb1_board
  GENERIC MAP (
    g_sim         => g_sim,
    g_design_name => g_design_name,
    g_stamp_date  => g_stamp_date,
    g_stamp_time  => g_stamp_time, 
    g_stamp_svn   => g_stamp_svn, 
    g_fw_version  => c_fw_version,
    g_mm_clk_freq => c_unb1_board_mm_clk_freq_50M,
    g_use_phy     => c_use_phy,
    g_aux         => c_unb1_board_aux,
    g_dp_clk_use_pll => FALSE -- Use altpll_1 in QSYS
  )
  PORT MAP (
    -- Clock an reset signals
    cs_sim                   => cs_sim,
    xo_clk                   => xo_clk,
    xo_rst                   => xo_rst,
    xo_rst_n                 => xo_rst_n,

    mm_clk                   => mm_clk,
    mm_locked                => mm_locked,
    mm_rst                   => mm_rst,

    epcs_clk                 => '0',
    
    dp_rst                   => dp_rst,
    dp_clk                   => OPEN, --dp_clk,
    dp_pps                   => OPEN,
    dp_rst_in                => dp_rst,
    dp_clk_in                => dp_clk,
    
    -- Toggle WDI
    pout_wdi                 => pout_wdi,

    -- MM buses
    -- . Manual WDI override
    reg_wdi_mosi             => reg_wdi_mosi,
    reg_wdi_miso             => reg_wdi_miso,
    
    -- . System_info
    reg_unb_system_info_mosi => reg_unb_system_info_mosi,
    reg_unb_system_info_miso => reg_unb_system_info_miso, 
    rom_unb_system_info_mosi => rom_unb_system_info_mosi,
    rom_unb_system_info_miso => rom_unb_system_info_miso, 
    
    -- . UniBoard I2C sensors
    reg_unb_sens_mosi        => reg_unb_sens_mosi,
    reg_unb_sens_miso        => reg_unb_sens_miso,    
    
    -- . PPSH
    reg_ppsh_mosi            => reg_ppsh_mosi,
    reg_ppsh_miso            => reg_ppsh_miso,

    -- eth1g
    eth1g_tse_clk            => eth1g_tse_clk,   -- 125 MHz from xo_clk PLL in SOPC system
    eth1g_mm_rst             => eth1g_mm_rst,
    eth1g_tse_mosi           => eth1g_tse_mosi,
    eth1g_tse_miso           => eth1g_tse_miso,
    eth1g_reg_mosi           => eth1g_reg_mosi,
    eth1g_reg_miso           => eth1g_reg_miso,
    eth1g_reg_interrupt      => eth1g_reg_interrupt,
    eth1g_ram_mosi           => eth1g_ram_mosi,
    eth1g_ram_miso           => eth1g_ram_miso,
        
    -- FPGA pins
    -- . General
    CLK                      => '0',
    PPS                      => PPS,
    WDI                      => WDI,
    INTA                     => INTA,
    INTB                     => INTB,
    -- . Others
    VERSION                  => VERSION,
    ID                       => ID,
    TESTIO                   => TESTIO,
    -- . I2C Interface to Sensors
    sens_sc                  => sens_sc,
    sens_sd                  => sens_sd,        
    -- . 1GbE Control Interface
    ETH_clk                  => ETH_clk,
    ETH_SGIN                 => ETH_SGIN,
    ETH_SGOUT                => ETH_SGOUT
  );

  -----------------------------------------------------------------------------
  -- MM master
  -----------------------------------------------------------------------------
  u_mmm : ENTITY work.mmm_apertif_unb1_correlator
  GENERIC MAP (
    g_sim         => g_sim,
    g_sim_unb_nr  => g_sim_unb_nr,
    g_sim_node_nr => g_sim_node_nr,
    g_wpfb        => c_wpfb   
   )
  PORT MAP(  
    xo_clk                   => xo_clk,       
    xo_rst_n                 => xo_rst_n,     
    xo_rst                   => xo_rst,       

    mm_rst                   => mm_rst,
    mm_clk                   => mm_clk,       
    mm_locked                => mm_locked,    

    clk_clk                  => CLK, -- altpll_1 ref clk (200MHz)
    clk_clk_in_reset_reset_n => xo_rst_n,

    dp_clk                   => dp_clk, -- altpll_1 output clock

    -- PIOs
    pout_wdi                 => pout_wdi,

    -- Manual WDI override
    reg_wdi_mosi             => reg_wdi_mosi,
    reg_wdi_miso             => reg_wdi_miso,

    -- system_info
    reg_unb_system_info_mosi => reg_unb_system_info_mosi,
    reg_unb_system_info_miso => reg_unb_system_info_miso,
    rom_unb_system_info_mosi => rom_unb_system_info_mosi,
    rom_unb_system_info_miso => rom_unb_system_info_miso, 

    -- UniBoard I2C sensors
    reg_unb_sens_mosi        => reg_unb_sens_mosi,
    reg_unb_sens_miso        => reg_unb_sens_miso, 
 
    -- PPSH
    reg_ppsh_mosi            => reg_ppsh_mosi,
    reg_ppsh_miso            => reg_ppsh_miso, 

    -- . Data buffers
    reg_diag_data_buf_mosi   => reg_diag_data_buf_mosi,
    reg_diag_data_buf_miso   => reg_diag_data_buf_miso,

    ram_diag_data_buf_mosi   => ram_diag_data_buf_mosi,
    ram_diag_data_buf_miso   => ram_diag_data_buf_miso,    

    -- Filter coefficients
    ram_fil_coefs_mosi       => ram_fil_coefs_mosi,
    ram_fil_coefs_miso       => ram_fil_coefs_miso,
  
    -- eth1g
    eth1g_tse_clk            => eth1g_tse_clk,
    eth1g_mm_rst             => eth1g_mm_rst,
    eth1g_tse_mosi           => eth1g_tse_mosi,
    eth1g_tse_miso           => eth1g_tse_miso,
    eth1g_reg_mosi           => eth1g_reg_mosi,
    eth1g_reg_miso           => eth1g_reg_miso,
    eth1g_reg_interrupt      => eth1g_reg_interrupt,
    eth1g_ram_mosi           => eth1g_ram_mosi,
    eth1g_ram_miso           => eth1g_ram_miso
  );

END str;

