-------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose: Verify the terminals for the UniBoard mesh and the backplane using
--          a block generator (BG) in the BN and a data buffer (DB) in the FN.
-- Description:
--   g_unb_sys:
--     unb  fn  bn
--     1    1   1     For fastest simulation to verify link and BSN align
--     1    1   4     Realistic simulation of one UNB with one FN
--     1    4   1     No added value compared to 1 FN
--     1    4   4     Complete UNB, functionally no added value compared to
--                    using 4 BN --> 1 FN, but it does accurately simulate the
--                    mesh PCB reorder.
--     4    1   4     Realistic simulation of a subrack with a back PCB and
--                    four UNB each with one FN
--     4    4   4     Complete subrack with four UNB with all BN and FN.
--
-- Usage:
--   Modelsim:
--   > as 8
--   . Unfold and observe:
--     - bg_sosi_arr in BN to see the streams that are generated by the BG
--     - back_*_usr_sosi_arr2 to see what is send between corresponding BN
--     - mesh_tx_usr_sosi_arr2 to see what is send to the FN via the mesh
--     - db_in_sosi_arr in FN after BSN aligner to see the aligned streams
--     - rx_usr_sosi_arr in FN to see what is received before the BSN aligner.
--
--   > as 15
--   . For observing FIFO usedw and uth_rx timeout_cnt
--
--   > run 30 us
--   . After issueing the run command start the Python test case in the shell.
--
--   > do tb_mmf_node_fn_terminal_db.do
--   . To verify rx_usr_sosi_2arr recovery after fault on unb[0], fn[0],
--     bus[0], lane [0], this can cause expected "Warning: uth_rx(rtl_hold)
--     timeout occurred!" dependent on when exactly the
--     tb_mmf_node_fn_terminal_db.do is issued.

--   > do tb_mmf_node_fn_terminal_db.do
--     . Try again (edit the do file to affect lane 0 on back or mesh)
--
--   Shell:
--   > cd $UPE
--
--   . For example do with g_unb_sys = (1, 1, 1):
--   > python apps/bn_bg_terminal_fn_db/tc_bn_bg_terminal_fn_db.py --unb 0 --bn 0 --fn 0 --sim
--
--   . or do with g_unb_sys = (4, 4, 4):
--   > python apps/bn_bg_terminal_fn_db/tc_bn_bg_terminal_fn_db.py --unb 0:3 --bn 0:3 --fn 0:3 --sim
--
--   . or do with g_unb_sys = (4, 1, 4):
--   > python apps/bn_bg_terminal_fn_db/tc_bn_bg_terminal_fn_db.py --unb 0:3 --bn 0:3 --fn 0 --sim
--

LIBRARY IEEE, common_lib, unb_common_lib, bn_terminal_bg_lib, fn_terminal_db_lib, mm_lib, bf_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.tb_common_pkg.ALL;
USE common_lib.tb_common_mem_pkg.ALL;
USE unb_common_lib.unb_common_pkg.ALL;
USE unb_common_lib.tb_unb_common_pkg.ALL;
USE bf_lib.bf_pkg.ALL;
USE mm_lib.mm_file_pkg.ALL;
USE mm_lib.mm_file_unb_pkg.ALL;

ENTITY tb_mmf_node_fn_terminal_db IS
  GENERIC (
    g_sim_level   : NATURAL := 1;           -- 0 = use accurate GX IP model, 1 = use fast GX behavioural model in tr_nonbonded
    g_unb_sys     : t_c_mmf_unb_sys := (4, 1, 4)  -- (nof_unb, nof_fn per UniBoard, nof_bn per UniBoard); 
  );
END tb_mmf_node_fn_terminal_db;

ARCHITECTURE tb OF tb_mmf_node_fn_terminal_db IS

  CONSTANT c_sim                       : BOOLEAN := TRUE;

  CONSTANT c_use_back                  : BOOLEAN := sel_a_b(g_unb_sys.nof_unb=4, TRUE, FALSE);  -- To interconnect multiple boards via the backplane when g_unb_sys.nof_unb=4 else when g_unb_sys.nof_unb=1
                                                                                                -- this loops back each back node's BN_BI_TX to BN_BI_RX.  
  CONSTANT c_ena_mesh_reorder          : BOOLEAN := TRUE;
  CONSTANT c_mesh_use_bidir            : BOOLEAN := FALSE;
  CONSTANT c_mesh_nof_serial           : NATURAL := 3;     -- default only use the 3 full featured transceivers in the mesh (out of maximal 4), using only 2 is not enough

  ----------------------------------------------------------------------------
  -- Clocks and resets
  ----------------------------------------------------------------------------   
  CONSTANT c_mm_clk_period             : TIME := 100 ps;
  CONSTANT c_dp_clk_period             : TIME := 5 ns;
  CONSTANT c_tr_clk_period             : TIME := 6.4 ns;
  CONSTANT c_cal_rec_clk_period        : TIME := 25 ns;
  CONSTANT c_dp_pps_period             : NATURAL := 8;

  SIGNAL dp_pps                        : STD_LOGIC;

  SIGNAL mm_rst                        : STD_LOGIC;
  SIGNAL mm_clk                        : STD_LOGIC := '1';

  SIGNAL dp_rst                        : STD_LOGIC;
  SIGNAL dp_clk                        : STD_LOGIC := '1';

  SIGNAL tr_clk                        : STD_LOGIC := '1';
  SIGNAL cal_rec_clk                   : STD_LOGIC := '1';

  TYPE t_mem_mosi_fn_2arr IS ARRAY  (g_unb_sys.nof_unb-1 DOWNTO 0) OF t_mem_mosi_arr(g_unb_sys.nof_fn-1 DOWNTO 0);
  TYPE t_mem_miso_fn_2arr IS ARRAY  (g_unb_sys.nof_unb-1 DOWNTO 0) OF t_mem_miso_arr(g_unb_sys.nof_fn-1 DOWNTO 0);

  TYPE t_mem_mosi_bn_2arr IS ARRAY  (g_unb_sys.nof_unb-1 DOWNTO 0) OF t_mem_mosi_arr(g_unb_sys.nof_bn-1 DOWNTO 0);
  TYPE t_mem_miso_bn_2arr IS ARRAY  (g_unb_sys.nof_unb-1 DOWNTO 0) OF t_mem_miso_arr(g_unb_sys.nof_bn-1 DOWNTO 0);

  ----------------------------------------------------------------------------
  -- MM buses for node_bn_terminal_bg
  ----------------------------------------------------------------------------                                         
  SIGNAL bn_reg_diag_bg_mosi_2arr            : t_mem_mosi_bn_2arr;
  SIGNAL bn_reg_diag_bg_miso_2arr            : t_mem_miso_bn_2arr;
   
  SIGNAL bn_ram_diag_bg_mosi_2arr            : t_mem_mosi_bn_2arr;
  SIGNAL bn_ram_diag_bg_miso_2arr            : t_mem_miso_bn_2arr;
   
  SIGNAL bn_reg_tr_nonbonded_mesh_mosi_2arr  : t_mem_mosi_bn_2arr;
  SIGNAL bn_reg_tr_nonbonded_mesh_miso_2arr  : t_mem_miso_bn_2arr;
   
  SIGNAL bn_reg_diagnostics_mesh_mosi_2arr   : t_mem_mosi_bn_2arr;
  SIGNAL bn_reg_diagnostics_mesh_miso_2arr   : t_mem_miso_bn_2arr;
   
  SIGNAL bn_reg_tr_nonbonded_back_mosi_2arr  : t_mem_mosi_bn_2arr;
  SIGNAL bn_reg_tr_nonbonded_back_miso_2arr  : t_mem_miso_bn_2arr;
   
  SIGNAL bn_reg_diagnostics_back_mosi_2arr   : t_mem_mosi_bn_2arr;
  SIGNAL bn_reg_diagnostics_back_miso_2arr   : t_mem_miso_bn_2arr;
   
  SIGNAL bn_ram_diag_data_buf_mesh_mosi_2arr : t_mem_mosi_bn_2arr;
  SIGNAL bn_ram_diag_data_buf_mesh_miso_2arr : t_mem_miso_bn_2arr;

  SIGNAL bn_reg_ppsh_mosi_2arr               : t_mem_mosi_bn_2arr;
  SIGNAL bn_reg_ppsh_miso_2arr               : t_mem_miso_bn_2arr;

  ----------------------------------------------------------------------------
  -- MM buses for node_fn_terminal_db
  ----------------------------------------------------------------------------   
  SIGNAL fn_reg_tr_nonbonded_mosi_2arr       : t_mem_mosi_fn_2arr;
  SIGNAL fn_reg_tr_nonbonded_miso_2arr       : t_mem_miso_fn_2arr;
   
  SIGNAL fn_reg_diagnostics_mosi_2arr        : t_mem_mosi_fn_2arr;
  SIGNAL fn_reg_diagnostics_miso_2arr        : t_mem_miso_fn_2arr;

  SIGNAL fn_ram_diag_data_buf_mosi_2arr      : t_mem_mosi_fn_2arr;
  SIGNAL fn_ram_diag_data_buf_miso_2arr      : t_mem_miso_fn_2arr;

  SIGNAL fn_reg_diag_data_buf_mosi_2arr      : t_mem_mosi_fn_2arr;
  SIGNAL fn_reg_diag_data_buf_miso_2arr      : t_mem_miso_fn_2arr;
   
  SIGNAL fn_ram_mesh_diag_data_buf_mosi_2arr : t_mem_mosi_fn_2arr;
  SIGNAL fn_ram_mesh_diag_data_buf_miso_2arr : t_mem_miso_fn_2arr;
   
  SIGNAL fn_reg_bsn_monitor_mosi_2arr        : t_mem_mosi_fn_2arr;
  SIGNAL fn_reg_bsn_monitor_miso_2arr        : t_mem_miso_fn_2arr;

  SIGNAL fn_reg_ppsh_mosi_2arr               : t_mem_mosi_fn_2arr;
  SIGNAL fn_reg_ppsh_miso_2arr               : t_mem_miso_fn_2arr;

  ------------------------------------------------------------------------------
  -- BN back side serial I/O
  ------------------------------------------------------------------------------
  SIGNAL bn_out_back_serial_4arr : t_unb_back_sl_4arr; 
  SIGNAL bn_in_back_serial_4arr  : t_unb_back_sl_4arr;

  ------------------------------------------------------------------------------
  -- BN & FN mesh side serial I/O
  ------------------------------------------------------------------------------
  SIGNAL bn_out_mesh_serial_4arr : t_unb_mesh_sl_4arr;
  SIGNAL bn_in_mesh_serial_4arr  : t_unb_mesh_sl_4arr;

  SIGNAL fn_in_mesh_serial_4arr  : t_unb_mesh_sl_4arr;
  SIGNAL fn_out_mesh_serial_4arr : t_unb_mesh_sl_4arr;
 
  ----------------------------------------------------------------------------
  -- Component declaration of mm_file (many instances in this TB)
  ----------------------------------------------------------------------------
  COMPONENT mm_file
  GENERIC(
    g_file_prefix       : STRING;
    g_mm_clk_period     : TIME := c_mm_clk_period;
    g_update_on_change  : BOOLEAN := FALSE
  );
  PORT (
    mm_rst        : IN  STD_LOGIC;
    mm_clk        : IN  STD_LOGIC;
    mm_master_out : OUT t_mem_mosi;
    mm_master_in  : IN  t_mem_miso 
  );
  END COMPONENT;

BEGIN

  ----------------------------------------------------------------------------
  -- Clock and reset generation
  ----------------------------------------------------------------------------
  mm_clk <= NOT mm_clk AFTER c_mm_clk_period/2;
  mm_rst <= '1', '0' AFTER c_mm_clk_period*5;

  dp_clk <= NOT dp_clk AFTER c_dp_clk_period/2;
  dp_rst <= '1', '0' AFTER c_dp_clk_period*5;

  tr_clk <= NOT tr_clk AFTER c_tr_clk_period/2;
  cal_rec_clk <= NOT cal_rec_clk AFTER c_cal_rec_clk_period/2;

  ----------------------------------------------------------------------------
  -- Procedure that polls a sim control file that can be used to e.g. get
  -- the simulation time in ns
  ----------------------------------------------------------------------------
  mmf_poll_sim_ctrl_file(c_mmf_unb_file_path & "sim.ctrl", c_mmf_unb_file_path & "sim.stat");
 
  ------------------------------------------------------------------------------
  -- External PPS
  ------------------------------------------------------------------------------  
  proc_common_gen_pulse(1, c_dp_pps_period, '1', dp_rst, dp_clk, dp_pps);

  ----------------------------------------------------------------------------
  -- DUTs and their MM buses
  ----------------------------------------------------------------------------
  gen_unb : FOR UNB IN 0 TO g_unb_sys.nof_unb-1 GENERATE  
    gen_bn: FOR BN IN 0 TO g_unb_sys.nof_bn-1 GENERATE  
      ----------------------------------------------------------------------------
      -- bn_terminal_bg: MM <-> file I/O
      ----------------------------------------------------------------------------
      --u_mm_file_reg_diagnostics_back   : mm_file GENERIC MAP(mmf_unb_file_prefix(UNB, BN, "BN") & "REG_DIAGNOSTICS_BACK") 
      --                                              PORT MAP(mm_rst, mm_clk, bn_reg_diagnostics_back_mosi_2arr(UNB)(BN), bn_reg_diagnostics_back_miso_2arr(UNB)(BN) );

      u_mm_file_reg_diag_bg            : mm_file GENERIC MAP(mmf_unb_file_prefix(UNB, BN, "BN") & "REG_DIAG_BG")
                                                    PORT MAP(mm_rst, mm_clk, bn_reg_diag_bg_mosi_2arr(UNB)(BN), bn_reg_diag_bg_miso_2arr(UNB)(BN) );

      u_mm_file_ram_diag_bg            : mm_file GENERIC MAP(mmf_unb_file_prefix(UNB, BN, "BN") & "RAM_DIAG_BG")
                                                    PORT MAP(mm_rst, mm_clk, bn_ram_diag_bg_mosi_2arr(UNB)(BN), bn_ram_diag_bg_miso_2arr(UNB)(BN) );
  
      --u_mm_file_reg_tr_nonbonded_mesh  : mm_file GENERIC MAP(mmf_unb_file_prefix(UNB, BN, "BN") & "REG_TR_NONBONDED_MESH")                
      --                                              PORT MAP(mm_rst, mm_clk, bn_reg_tr_nonbonded_mesh_mosi_2arr(UNB)(BN), bn_reg_tr_nonbonded_mesh_miso_2arr(UNB)(BN) );
 
      --u_mm_file_reg_diagnostics_mesh   : mm_file GENERIC MAP(mmf_unb_file_prefix(UNB, BN, "BN") & "REG_DIAGNOSTICS_MESH")                    
      --                                              PORT MAP(mm_rst, mm_clk, bn_reg_diagnostics_mesh_mosi_2arr(UNB)(BN), bn_reg_diagnostics_mesh_miso_2arr(UNB)(BN) );

      --u_mm_file_reg_tr_nonbonded_back  : mm_file GENERIC MAP(mmf_unb_file_prefix(UNB, BN, "BN") & "REG_TR_NONBONDED_BACK")                   
      --                                              PORT MAP(mm_rst, mm_clk, bn_reg_tr_nonbonded_back_mosi_2arr(UNB)(BN), bn_reg_tr_nonbonded_back_miso_2arr(UNB)(BN) );

      --u_mm_file_ram_diag_data_buf_mesh : mm_file GENERIC MAP(mmf_unb_file_prefix(UNB, BN, "BN") & "RAM_DIAG_DATA_BUF_MESH")
      --                                              PORT MAP(mm_rst, mm_clk, bn_ram_diag_data_buf_mesh_mosi_2arr(UNB)(BN), bn_ram_diag_data_buf_mesh_miso_2arr(UNB)(BN) );

      --u_mm_file_reg_ppsh               : mm_file GENERIC MAP(mmf_unb_file_prefix(UNB, BN, "BN") & "PIO_PPS") -- Using the old PIO name
      --                                              PORT MAP(mm_rst, mm_clk, bn_reg_ppsh_mosi_2arr(UNB)(BN), bn_reg_ppsh_miso_2arr(UNB)(BN) );
      
      ----------------------------------------------------------------------------
      -- bn_terminal_bg: Node function: block generator & terminals
      ---------------------------------------------------------------------------- 
      u_node_bn_terminal_bg : ENTITY bn_terminal_bg_lib.node_bn_terminal_bg
      GENERIC MAP(
        g_sim                     => c_sim,
        g_sim_level               => g_sim_level,
        g_use_back                => c_use_back,
        g_mesh_nof_serial         => c_mesh_nof_serial,
        g_mesh_use_rx             => c_mesh_use_bidir,
        g_mesh_ena_reorder        => c_ena_mesh_reorder
      )
      PORT MAP(
        -- System
        mm_rst                      => mm_rst,
        mm_clk                      => mm_clk,
        dp_rst                      => dp_rst,
        dp_clk                      => dp_clk,
        dp_pps                      => dp_pps,
        tr_mesh_clk                 => tr_CLK,
        tr_back_clk                 => tr_CLK,
        cal_clk                     => cal_rec_clk,
        
        chip_id                     => TO_UVEC(BN +4, c_unb_nof_chip_w),   -- BN chip ID 4,5,6,7 
        bck_id                      => TO_UVEC(UNB, c_unb_nof_uniboard_w), -- Backplane ID 0,1,2,3 
    
        -- MM interface
        -- . block generator
        reg_diag_bg_mosi            => bn_reg_diag_bg_mosi_2arr(UNB)(BN),
        reg_diag_bg_miso            => bn_reg_diag_bg_miso_2arr(UNB)(BN),
        ram_diag_bg_mosi            => bn_ram_diag_bg_mosi_2arr(UNB)(BN),
        ram_diag_bg_miso            => bn_ram_diag_bg_miso_2arr(UNB)(BN),
        -- . tr_nonbonded mesh
        reg_mesh_tr_nonbonded_mosi  => bn_reg_tr_nonbonded_mesh_mosi_2arr(UNB)(BN),
        reg_mesh_tr_nonbonded_miso  => bn_reg_tr_nonbonded_mesh_miso_2arr(UNB)(BN),
        reg_mesh_diagnostics_mosi   => bn_reg_diagnostics_mesh_mosi_2arr(UNB)(BN),
        reg_mesh_diagnostics_miso   => bn_reg_diagnostics_mesh_miso_2arr(UNB)(BN),    
        -- . tr_nonbonded back
        reg_back_tr_nonbonded_mosi  => bn_reg_tr_nonbonded_back_mosi_2arr(UNB)(BN),
        reg_back_tr_nonbonded_miso  => bn_reg_tr_nonbonded_back_miso_2arr(UNB)(BN),
        reg_back_diagnostics_mosi   => bn_reg_diagnostics_back_mosi_2arr(UNB)(BN),
        reg_back_diagnostics_miso   => bn_reg_diagnostics_back_miso_2arr(UNB)(BN),    
        -- . diag_data_buffer mesh
        ram_mesh_diag_data_buf_mosi => bn_ram_diag_data_buf_mesh_mosi_2arr(UNB)(BN),
        ram_mesh_diag_data_buf_miso => bn_ram_diag_data_buf_mesh_miso_2arr(UNB)(BN),
       
        -- Mesh interface level
        -- . Serial (tr_nonbonded)
        mesh_tx_serial_2arr         => bn_out_mesh_serial_4arr(UNB)(BN),
        mesh_rx_serial_2arr         => bn_in_mesh_serial_4arr(UNB)(BN),

        -- Back interface level
        -- . Serial (tr_nonbonded)
        back_tx_serial_2arr         => bn_out_back_serial_4arr(UNB)(BN),
        back_rx_serial_2arr         => bn_in_back_serial_4arr(UNB)(BN)
      );
 
    END GENERATE;

    gen_fn: FOR FN IN 0 TO g_unb_sys.nof_fn-1 GENERATE  
      ----------------------------------------------------------------------------
      -- fn_terminal_db: MM <-> file I/O
      ----------------------------------------------------------------------------
      --u_mm_file_reg_tr_nonbonded       : mm_file GENERIC MAP(mmf_unb_file_prefix(UNB, FN, "FN") & "REG_TR_NONBONDED")
      --                                              PORT MAP(mm_rst, mm_clk, fn_reg_tr_nonbonded_mosi_2arr(UNB)(FN), fn_reg_tr_nonbonded_miso_2arr(UNB)(FN) );

      --u_mm_file_reg_diagnostics        : mm_file GENERIC MAP(mmf_unb_file_prefix(UNB, FN, "FN") & "REG_DIAGNOSTICS_MESH")
      --                                              PORT MAP(mm_rst, mm_clk, fn_reg_diagnostics_mosi_2arr(UNB)(FN), fn_reg_diagnostics_miso_2arr(UNB)(FN) );

      u_mm_file_ram_diag_data_buf      : mm_file GENERIC MAP(mmf_unb_file_prefix(UNB, FN, "FN") & "RAM_DIAG_DATA_BUFFER")
                                                    PORT MAP(mm_rst, mm_clk, fn_ram_diag_data_buf_mosi_2arr(UNB)(FN), fn_ram_diag_data_buf_miso_2arr(UNB)(FN) );

      u_mm_file_reg_diag_data_buf      : mm_file GENERIC MAP(mmf_unb_file_prefix(UNB, FN, "FN") & "REG_DIAG_DATA_BUFFER")
                                                    PORT MAP(mm_rst, mm_clk, fn_reg_diag_data_buf_mosi_2arr(UNB)(FN), fn_reg_diag_data_buf_miso_2arr(UNB)(FN) );

      --u_mm_file_ram_diag_data_buf_mesh : mm_file GENERIC MAP(mmf_unb_file_prefix(UNB, FN, "FN") & "RAM_DIAG_DATA_BUFFER_MESH")
      --                                              PORT MAP(mm_rst, mm_clk, fn_ram_mesh_diag_data_buf_mosi_2arr(UNB)(FN), fn_ram_mesh_diag_data_buf_miso_2arr(UNB)(FN) );

      --u_mm_file_reg_bsn_monitor        : mm_file GENERIC MAP(mmf_unb_file_prefix(UNB, FN, "FN") & "REG_BSN_MONITOR")
      --                                              PORT MAP(mm_rst, mm_clk, fn_reg_bsn_monitor_mosi_2arr(UNB)(FN), fn_reg_bsn_monitor_miso_2arr(UNB)(FN) );

      --u_mm_file_reg_ppsh               : mm_file GENERIC MAP(mmf_unb_file_prefix(UNB, FN, "FN") & "PIO_PPS") -- Using the old PIO name
      --                                              PORT MAP(mm_rst, mm_clk, fn_reg_ppsh_mosi_2arr(UNB)(FN), fn_reg_ppsh_miso_2arr(UNB)(FN) );

      -----------------------------------------------------------------------------
      -- Node function: Terminals and data buffer
      -----------------------------------------------------------------------------  
      u_node_fn_terminal_db : ENTITY fn_terminal_db_lib.node_fn_terminal_db
      GENERIC MAP(
        g_sim                       => c_sim,
        g_sim_level                 => g_sim_level,
        g_use_bsn_align             => TRUE,
        g_use_data_buf              => TRUE,
        -- Terminals interface
        g_multi_unb                 => sel_a_b(g_unb_sys.nof_unb>1, TRUE, FALSE),
        g_mesh_nof_serial           => c_mesh_nof_serial,
        g_mesh_use_tx               => c_mesh_use_bidir,
        g_mesh_ena_reorder          => c_ena_mesh_reorder
      )
      PORT MAP(
        -- System
        mm_rst                      => mm_rst,
        mm_clk                      => mm_clk,
        dp_rst                      => dp_rst,
        dp_clk                      => dp_clk,
        dp_pps                      => dp_pps,
        tr_mesh_clk                 => tr_clk,
        cal_clk                     => cal_rec_clk,
    
        chip_id                     => TO_UVEC(FN, c_unb_nof_chip_w), -- FN chip ID 0,1,2,3
    
        -- MM interface
        -- . tr_nonbonded
        reg_tr_nonbonded_mosi       => fn_reg_tr_nonbonded_mosi_2arr(UNB)(FN),
        reg_tr_nonbonded_miso       => fn_reg_tr_nonbonded_miso_2arr(UNB)(FN),
        reg_diagnostics_mosi        => fn_reg_diagnostics_mosi_2arr(UNB)(FN),
        reg_diagnostics_miso        => fn_reg_diagnostics_miso_2arr(UNB)(FN),
        -- . diag_data_buffer
        ram_diag_data_buf_mosi      => fn_ram_diag_data_buf_mosi_2arr(UNB)(FN),
        ram_diag_data_buf_miso      => fn_ram_diag_data_buf_miso_2arr(UNB)(FN),
        reg_diag_data_buf_mosi      => fn_reg_diag_data_buf_mosi_2arr(UNB)(FN),
        reg_diag_data_buf_miso      => fn_reg_diag_data_buf_miso_2arr(UNB)(FN),
        -- . diag_data_buffer_mesh
        ram_mesh_diag_data_buf_mosi => fn_ram_mesh_diag_data_buf_mosi_2arr(UNB)(FN),
        ram_mesh_diag_data_buf_miso => fn_ram_mesh_diag_data_buf_miso_2arr(UNB)(FN),
        -- . bsn_monitor
        reg_bsn_monitor_mosi        => fn_reg_bsn_monitor_mosi_2arr(UNB)(FN),
        reg_bsn_monitor_miso        => fn_reg_bsn_monitor_miso_2arr(UNB)(FN),
  
        -- Mesh interface level
        -- . Serial (tr_nonbonded)
        tx_serial_2arr              => fn_out_mesh_serial_4arr(UNB)(FN),  -- Tx support for diagnostics
        rx_serial_2arr              => fn_in_mesh_serial_4arr(UNB)(FN)    -- Rx
      );
 
    END GENERATE;

    ------------------------------------------------------------------------------
    -- Interconnect BN --> FN for each UniBoard
    --
    -- . t_unb_mesh_sl_4arr indexing [unb][node][bus][lane]
    -- . t_unb_back_sl_4arr indexing [unb][  bn][bus][lane]
    ------------------------------------------------------------------------------
  
    -- Direct interconnect BN0<->FN0.
    no_mesh : IF g_unb_sys.nof_bn=1 AND g_unb_sys.nof_fn=1 GENERATE
      fn_in_mesh_serial_4arr(UNB)(0) <= bn_out_mesh_serial_4arr(UNB)(0);
      bn_in_mesh_serial_4arr(UNB)(0) <= fn_out_mesh_serial_4arr(UNB)(0);
    END GENERATE;
      
    -- Mesh model
    gen_mesh : IF g_unb_sys.nof_bn>1 OR g_unb_sys.nof_fn>1 GENERATE
      u_mesh_model_serial : ENTITY unb_common_lib.unb_mesh_model_sl
      GENERIC MAP(
        g_reorder      => c_ena_mesh_reorder
      )
      PORT MAP (
        -- FN to BN
        fn_tx_sl_3arr  => fn_out_mesh_serial_4arr(UNB),
        bn_rx_sl_3arr  => bn_in_mesh_serial_4arr(UNB),
        
        -- BN to FN
        bn_tx_sl_3arr  => bn_out_mesh_serial_4arr(UNB),
        fn_rx_sl_3arr  => fn_in_mesh_serial_4arr(UNB)
      );
    END GENERATE;
   
  END GENERATE;

   ------------------------------------------------------------------------------
   -- Instantiate a backplane model that interconnects all UniBoards...
   ------------------------------------------------------------------------------  
  gen_backplane : IF c_use_back=TRUE GENERATE
    gen_model : ENTITY unb_common_lib.unb_back_model_sl
    PORT MAP (
      backplane_in_serial_4arr  => bn_out_back_serial_4arr,
      backplane_out_serial_4arr => bn_in_back_serial_4arr
   );
  END GENERATE;

   ------------------------------------------------------------------------------
   -- ...or loop back serial TX to RX in case of a single UniBoard.
   ------------------------------------------------------------------------------  
  no_backplane: IF c_use_back=FALSE GENERATE
    bn_in_back_serial_4arr <= bn_out_back_serial_4arr;
  END GENERATE;

END tb;
