--------------------------------------------------------------------------------
--
-- Copyright (C) 2015
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

-- Purpose: Test bench for mms_diag_tx_seq --> mms_diag_rx_seq
--   The tb is self-stopping and self-checking.
-- Usage:
--   > as 5
--   > run -all


LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.STD_LOGIC_1164.ALL;    
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_lfsr_sequences_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.tb_common_mem_pkg.ALL;
USE common_lib.tb_common_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE dp_lib.tb_dp_pkg.ALL;
USE work.diag_pkg.ALL;
USE work.tb_diag_pkg.ALL;

ENTITY tb_mms_diag_seq IS
  GENERIC (   
    -- general
    g_flow_control_verify  : t_dp_flow_control_enum := e_active;  -- always active or random flow control
    -- specific
    g_nof_streams          : NATURAL := 1;
    g_mm_broadcast_tx      : BOOLEAN := FALSE;
    g_data_w               : NATURAL := 40;  -- >= g_seq_dat_w
    g_seq_dat_w            : NATURAL := 32
  );
END ENTITY tb_mms_diag_seq;

ARCHITECTURE str of tb_mms_diag_seq IS

  CONSTANT mm_clk_period          : TIME := 8 ns;    -- 125 MHz
  CONSTANT dp_clk_period          : TIME := 5 ns;    -- 200 MHz
  CONSTANT c_random_w             : NATURAL := 16;
  
  SIGNAL random                   : STD_LOGIC_VECTOR(c_random_w-1 DOWNTO 0) := (OTHERS=>'0');  -- use different lengths to have different random sequences
  SIGNAL ready                    : STD_LOGIC;
  
  SIGNAL tb_end                   : STD_LOGIC := '0';
  SIGNAL tb_mode                  : t_tb_diag_seq_mode_enum := s_off;
  SIGNAL tb_mode_multi            : t_tb_diag_seq_mode_enum := s_off;
  SIGNAL tb_verify                : STD_LOGIC := '0';
  SIGNAL mm_rst                   : STD_LOGIC;
  SIGNAL mm_clk                   : STD_LOGIC := '0';
  SIGNAL dp_rst                   : STD_LOGIC := '1';
  SIGNAL dp_clk                   : STD_LOGIC := '0';

  SIGNAL reg_tx_mosi              : t_mem_mosi := c_mem_mosi_rst;
  SIGNAL reg_tx_miso              : t_mem_miso;

  SIGNAL reg_rx_mosi              : t_mem_mosi := c_mem_mosi_rst;
  SIGNAL reg_rx_miso              : t_mem_miso;

  SIGNAL rd_reg_arr               : t_diag_seq_mm_reg_arr(g_nof_streams DOWNTO 0);  -- use +1 to support g_nof_streams=1, 2 in p_stimuli_and_verify

  SIGNAL tx_src_out_arr           : t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0);
  SIGNAL tx_src_in_arr            : t_dp_siso_arr(g_nof_streams-1 DOWNTO 0);

  SIGNAL stream_index             : NATURAL;
  SIGNAL force_low_error          : STD_LOGIC;
  SIGNAL force_replicate_error    : STD_LOGIC;
  
  SIGNAL rx_snk_in_arr            : t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0);
  
BEGIN

  mm_rst <= '1', '0' AFTER mm_clk_period * 5;
  dp_rst <= mm_rst WHEN rising_edge(dp_clk);

  mm_clk <= NOT mm_clk OR tb_end AFTER mm_clk_period/2;
  dp_clk <= NOT dp_clk OR tb_end AFTER dp_clk_period/2;
   
  ------------------------------------------------------------------------------
  -- STREAM CONTROL
  ------------------------------------------------------------------------------
  
  random <= func_common_random(random) WHEN rising_edge(dp_clk);
                       
  ready <= '1'                 WHEN g_flow_control_verify=e_active  ELSE
           random(random'HIGH) WHEN g_flow_control_verify=e_random  ELSE
           '1';
           
  tx_src_in_arr <= func_dp_stream_arr_set(tx_src_in_arr, ready, "READY");
                          
  ------------------------------------------------------------------------------
  -- Stimuli
  ------------------------------------------------------------------------------
  
  
  p_stimuli_and_verify : PROCESS
    -- Cannot use non static rd_reg_arr(I) variable index in procedure argument, therefor use constant index rd_reg_arr(c_st_0) or rd_reg_arr(c_st_1).
    CONSTANT c_st_0 : NATURAL := 0;
    CONSTANT c_st_1 : NATURAL := 1;
  BEGIN
    stream_index <= 0;   -- default verify stream 0
    force_low_error       <= '0';
    force_replicate_error <= '0';
    tb_mode <= s_off;
    proc_common_wait_until_low(mm_clk, mm_rst);
    proc_common_wait_some_cycles(mm_clk, 10);
    
    -------------------------------------------------------------------------
    -- Verify s_off
    -------------------------------------------------------------------------
    proc_diag_seq_verify(c_st_0, mm_clk, reg_tx_miso, reg_tx_mosi, reg_rx_miso, reg_rx_mosi, tb_mode, tb_verify, rd_reg_arr(c_st_0));
    IF g_nof_streams>1 THEN
      stream_index <= 1;   -- try stream 1
      proc_diag_seq_verify(c_st_1, mm_clk, reg_tx_miso, reg_tx_mosi, reg_rx_miso, reg_rx_mosi, tb_mode_multi, tb_verify, rd_reg_arr(c_st_1));
      stream_index <= 0;   -- back to stream 0
    END IF;
    
    -------------------------------------------------------------------------
    -- Verify Tx and Rx on and both with the same pattern
    -------------------------------------------------------------------------
    tb_mode <= s_expect_ok;
    proc_diag_seq_tx_disable(c_st_0, mm_clk, dp_clk, reg_tx_miso, reg_tx_mosi, reg_rx_miso, reg_rx_mosi, rd_reg_arr(c_st_0));
    proc_diag_seq_rx_disable(c_st_0, mm_clk, dp_clk, reg_tx_miso, reg_tx_mosi, reg_rx_miso, reg_rx_mosi, rd_reg_arr(c_st_0));
    proc_diag_seq_tx_enable(c_st_0, "CNTR", 17, mm_clk, dp_clk, reg_tx_miso, reg_tx_mosi, reg_rx_miso, reg_rx_mosi, rd_reg_arr(c_st_0));
    proc_diag_seq_rx_enable(c_st_0, "CNTR", mm_clk, dp_clk, reg_tx_miso, reg_tx_mosi, reg_rx_miso, reg_rx_mosi, rd_reg_arr(c_st_0));
    
    -- Run test and read and verify Rx status
    proc_common_wait_some_cycles(mm_clk, 200);
    proc_diag_seq_verify(c_st_0, mm_clk, reg_tx_miso, reg_tx_mosi, reg_rx_miso, reg_rx_mosi, tb_mode, tb_verify, rd_reg_arr(c_st_0));
    
    -- Check that there is valid data
    IF g_flow_control_verify=e_active THEN
      ASSERT rx_snk_in_arr(0).valid='1' REPORT "Wrong diag result: no valid data at rx_snk_in_arr(0)." SEVERITY ERROR;
    END IF;
      
    -------------------------------------------------------------------------
    -- Verify Tx and Rx on but with different pattern
    -------------------------------------------------------------------------
    tb_mode <= s_expect_error;
    proc_diag_seq_tx_enable(c_st_0, "PSRG", 17, mm_clk, dp_clk, reg_tx_miso, reg_tx_mosi, reg_rx_miso, reg_rx_mosi, rd_reg_arr(c_st_0));
    
    -- Run test and read and verify Rx status
    proc_common_wait_some_cycles(mm_clk, 200);
    proc_diag_seq_verify(c_st_0, mm_clk, reg_tx_miso, reg_tx_mosi, reg_rx_miso, reg_rx_mosi, tb_mode, tb_verify, rd_reg_arr(c_st_0));
          
    -------------------------------------------------------------------------
    -- Verify Rx off
    -------------------------------------------------------------------------
    tb_mode <= s_expect_no_result;
    proc_diag_seq_rx_disable(c_st_0, mm_clk, dp_clk, reg_tx_miso, reg_tx_mosi, reg_rx_miso, reg_rx_mosi, rd_reg_arr(c_st_0));
    
    -- Run test and read and verify Rx status
    proc_common_wait_some_cycles(mm_clk, 200);
    proc_diag_seq_verify(c_st_0, mm_clk, reg_tx_miso, reg_tx_mosi, reg_rx_miso, reg_rx_mosi, tb_mode, tb_verify, rd_reg_arr(c_st_0));
    
    IF g_mm_broadcast_tx=FALSE THEN
      -------------------------------------------------------------------------
      -- Verify Tx and Rx on with error in sequence low part
      -------------------------------------------------------------------------
      tb_mode <= s_expect_ok;
      proc_diag_seq_tx_disable(c_st_0, mm_clk, dp_clk, reg_tx_miso, reg_tx_mosi, reg_rx_miso, reg_rx_mosi, rd_reg_arr(c_st_0));
      proc_diag_seq_rx_disable(c_st_0, mm_clk, dp_clk, reg_tx_miso, reg_tx_mosi, reg_rx_miso, reg_rx_mosi, rd_reg_arr(c_st_0));
      proc_diag_seq_tx_enable(c_st_0, "CNTR", 17, mm_clk, dp_clk, reg_tx_miso, reg_tx_mosi, reg_rx_miso, reg_rx_mosi, rd_reg_arr(c_st_0));
      proc_diag_seq_rx_enable(c_st_0, "CNTR", mm_clk, dp_clk, reg_tx_miso, reg_tx_mosi, reg_rx_miso, reg_rx_mosi, rd_reg_arr(c_st_0));
      
      -- Run test and read and verify Rx status
      proc_common_wait_some_cycles(mm_clk, 200);
      proc_diag_seq_verify(c_st_0, mm_clk, reg_tx_miso, reg_tx_mosi, reg_rx_miso, reg_rx_mosi, tb_mode, tb_verify, rd_reg_arr(c_st_0));
      
      tb_mode <= s_expect_error;
      proc_common_wait_some_cycles(dp_clk, 1);
      force_low_error <= '1';
      proc_common_wait_some_cycles(dp_clk, c_random_w);  -- sufficently long to affect a valid data in case g_flow_control_verify=e_random
      force_low_error <= '0';
      proc_common_wait_some_cycles(mm_clk, 200);
      proc_diag_seq_verify(c_st_0, mm_clk, reg_tx_miso, reg_tx_mosi, reg_rx_miso, reg_rx_mosi, tb_mode, tb_verify, rd_reg_arr(c_st_0));
    END IF;
      
    IF g_mm_broadcast_tx=FALSE AND g_data_w>g_seq_dat_w THEN
      -------------------------------------------------------------------------
      -- Verify Tx and Rx on with error in sequence replicate part
      -------------------------------------------------------------------------
      tb_mode <= s_expect_ok;
      proc_diag_seq_tx_disable(c_st_0, mm_clk, dp_clk, reg_tx_miso, reg_tx_mosi, reg_rx_miso, reg_rx_mosi, rd_reg_arr(c_st_0));
      proc_diag_seq_rx_disable(c_st_0, mm_clk, dp_clk, reg_tx_miso, reg_tx_mosi, reg_rx_miso, reg_rx_mosi, rd_reg_arr(c_st_0));
      proc_diag_seq_tx_enable(c_st_0, "CNTR", 17, mm_clk, dp_clk, reg_tx_miso, reg_tx_mosi, reg_rx_miso, reg_rx_mosi, rd_reg_arr(c_st_0));
      proc_diag_seq_rx_enable(c_st_0, "CNTR", mm_clk, dp_clk, reg_tx_miso, reg_tx_mosi, reg_rx_miso, reg_rx_mosi, rd_reg_arr(c_st_0));
      
      -- Run test and read and verify Rx status
      proc_common_wait_some_cycles(mm_clk, 200);
      proc_diag_seq_verify(c_st_0, mm_clk, reg_tx_miso, reg_tx_mosi, reg_rx_miso, reg_rx_mosi, tb_mode, tb_verify, rd_reg_arr(c_st_0));
      
      tb_mode <= s_expect_error;
      proc_common_wait_some_cycles(dp_clk, 1);
      force_replicate_error <= '1';
      proc_common_wait_some_cycles(dp_clk, c_random_w);  -- sufficently long to affect a valid data in case g_flow_control_verify=e_random
      force_replicate_error <= '0';
      proc_common_wait_some_cycles(mm_clk, 200);
      proc_diag_seq_verify(c_st_0, mm_clk, reg_tx_miso, reg_tx_mosi, reg_rx_miso, reg_rx_mosi, tb_mode, tb_verify, rd_reg_arr(c_st_0));
    END IF;
      
    IF g_mm_broadcast_tx=FALSE AND g_nof_streams>1 THEN
      stream_index <= 1;   -- try stream 1
      -------------------------------------------------------------------------
      -- Verify Tx and Rx on and both with the same pattern for stream 1
      -------------------------------------------------------------------------
      proc_diag_seq_tx_disable(c_st_1, mm_clk, dp_clk, reg_tx_miso, reg_tx_mosi, reg_rx_miso, reg_rx_mosi, rd_reg_arr(c_st_1));
      proc_diag_seq_rx_disable(c_st_1, mm_clk, dp_clk, reg_tx_miso, reg_tx_mosi, reg_rx_miso, reg_rx_mosi, rd_reg_arr(c_st_1));
      proc_diag_seq_tx_enable(c_st_1, "CNTR", 17, mm_clk, dp_clk, reg_tx_miso, reg_tx_mosi, reg_rx_miso, reg_rx_mosi, rd_reg_arr(c_st_1));
      proc_diag_seq_rx_enable(c_st_1, "CNTR", mm_clk, dp_clk, reg_tx_miso, reg_tx_mosi, reg_rx_miso, reg_rx_mosi, rd_reg_arr(c_st_1));
      
      -- Run test and read and verify Rx status
      proc_common_wait_some_cycles(mm_clk, 200);
      tb_mode_multi <= s_expect_ok;
      proc_diag_seq_verify(c_st_1, mm_clk, reg_tx_miso, reg_tx_mosi, reg_rx_miso, reg_rx_mosi, tb_mode_multi, tb_verify, rd_reg_arr(c_st_1));
      
      -- Check that there is valid data
      IF g_flow_control_verify=e_active THEN
        ASSERT rx_snk_in_arr(1).valid='1' REPORT "Wrong diag result: no valid data at rx_snk_in_arr(1)." SEVERITY ERROR;
      END IF;
      stream_index <= 0;   -- back to stream 0
    END IF;
    
    -------------------------------------------------------------------------
    -- Both off
    -------------------------------------------------------------------------
    tb_mode <= s_off;
    proc_diag_seq_tx_disable(c_st_0, mm_clk, dp_clk, reg_tx_miso, reg_tx_mosi, reg_rx_miso, reg_rx_mosi, rd_reg_arr(c_st_0));
    proc_diag_seq_rx_disable(c_st_0, mm_clk, dp_clk, reg_tx_miso, reg_tx_mosi, reg_rx_miso, reg_rx_mosi, rd_reg_arr(c_st_0));
    proc_common_wait_some_cycles(mm_clk, 100);
    proc_diag_seq_verify(c_st_0, mm_clk, reg_tx_miso, reg_tx_mosi, reg_rx_miso, reg_rx_mosi, tb_mode, tb_verify, rd_reg_arr(c_st_0));
    proc_common_wait_some_cycles(mm_clk, 10);
    tb_end <= '1';
    WAIT;
  END PROCESS;

  p_verify_mm_broadcast : PROCESS(rx_snk_in_arr)
    VARIABLE v_snk_in : t_dp_sosi;
  BEGIN
    IF g_mm_broadcast_tx=TRUE AND g_nof_streams>1 THEN
      v_snk_in := rx_snk_in_arr(0);
      FOR I IN 1 TO g_nof_streams-1 LOOP
        IF v_snk_in/=rx_snk_in_arr(I) THEN
          REPORT "Wrong diag result: for g_mm_broadcast_tx=TRUE all streams should carry the same data." SEVERITY ERROR;
          EXIT;
        END IF;
      END LOOP;
    END IF;
  END PROCESS;
        
  
  u_mms_diag_tx_seq: ENTITY WORK.mms_diag_tx_seq
  GENERIC MAP(
    g_mm_broadcast => g_mm_broadcast_tx,
    g_nof_streams  => g_nof_streams,
    g_seq_dat_w    => g_seq_dat_w
  )
  PORT MAP(
    -- Clocks and reset
    mm_rst         => mm_rst,
    mm_clk         => mm_clk,
    dp_rst         => dp_rst,
    dp_clk         => dp_clk,

    -- MM interface
    reg_mosi       => reg_tx_mosi,
    reg_miso       => reg_tx_miso,

    -- DP streaming interface
    tx_src_out_arr => tx_src_out_arr,
    tx_src_in_arr  => tx_src_in_arr
  );      

  u_mms_diag_rx_seq: ENTITY WORK.mms_diag_rx_seq
  GENERIC MAP(
    g_nof_streams => g_nof_streams,
    g_seq_dat_w   => g_seq_dat_w,
    g_data_w      => g_data_w
  )
  PORT MAP(
    -- Clocks and reset
    mm_rst         => mm_rst,
    mm_clk         => mm_clk,
    dp_rst         => dp_rst,
    dp_clk         => dp_clk,

    -- MM interface
    reg_mosi       => reg_rx_mosi,
    reg_miso       => reg_rx_miso,

    -- DP streaming interface
    rx_snk_in_arr  => rx_snk_in_arr
  );      

  p_connect : PROCESS(tx_src_out_arr, force_low_error, force_replicate_error)
  BEGIN
    -- Default lopoback all streams
    rx_snk_in_arr <= tx_src_out_arr;
    -- Optionally apply errors on stream 0
    IF force_low_error='1' THEN 
      rx_snk_in_arr(0).data(0) <= NOT tx_src_out_arr(0).data(0);
    END IF;
    IF force_replicate_error='1' THEN 
      rx_snk_in_arr(0).data(g_seq_dat_w) <= NOT tx_src_out_arr(0).data(g_seq_dat_w);
    END IF;
  END PROCESS;
  
END ARCHITECTURE str;
