
-------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------
--
-- Purpose:  Testbench for the reorder_transpose unit 
--           To be used in conjunction with python testscript: ../python/tc_reorder_transpose.py
--
--
-- Usage:
--   > as 8
--   > run -all
--   > run python script in separate terminal: "python tc_reorder_transpose.py --unb 0 --fn 0 --sim"
--   > Stop the simulation manually in Modelsim by pressing the stop-button.
--   > Evalute the WAVE window. 

LIBRARY IEEE, common_lib, mm_lib, diag_lib, dp_lib, io_ddr_lib, technology_lib, tech_ddr_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.common_str_pkg.ALL;
USE common_lib.tb_common_pkg.ALL;
USE common_lib.tb_common_mem_pkg.ALL;
USE mm_lib.mm_file_unb_pkg.ALL;
USE mm_lib.mm_file_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL; 
USE diag_lib.diag_pkg.ALL;
USE technology_lib.technology_select_pkg.ALL;
USE tech_ddr_lib.tech_ddr_pkg.ALL;
USE work.reorder_pkg.ALL;


ENTITY tb_reorder_transpose IS 
  GENERIC (
    g_wr_chunksize     : POSITIVE := 256;
    g_rd_chunksize     : POSITIVE := 16;
    g_rd_nof_chunks    : POSITIVE := 16;  
    g_rd_interval      : POSITIVE := 16;   -- When pre-transpose is used equal to g_rd_chunksize. 
    g_gapsize          : NATURAL  := 0;
    g_nof_blocks       : POSITIVE := 32;
    g_nof_streams      : POSITIVE := 4;
    g_in_dat_w         : POSITIVE := 8;    
    g_mem_dat_w        : NATURAL  := 256;       -- The data width to the attached memory. 
    g_ena_pre_transp   : BOOLEAN  := TRUE
 );
END tb_reorder_transpose;

ARCHITECTURE tb OF tb_reorder_transpose IS
  
  CONSTANT c_sim                : BOOLEAN := TRUE;

  ----------------------------------------------------------------------------
  -- Clocks and resets
  ----------------------------------------------------------------------------   
  CONSTANT c_mm_clk_period      : TIME := 100 ps;
  CONSTANT c_dp_clk_period      : TIME := 5 ns;
  CONSTANT c_dp_pps_period      : NATURAL := 64;

  SIGNAL dp_pps                 : STD_LOGIC;

  SIGNAL mm_rst                 : STD_LOGIC := '1';
  SIGNAL mm_clk                 : STD_LOGIC := '0';

  SIGNAL dp_rst                 : STD_LOGIC;
  SIGNAL dp_clk                 : STD_LOGIC := '0';
  SIGNAL dp_rst_in              : STD_LOGIC;
  SIGNAL dp_clk_in              : STD_LOGIC := '0';

  ----------------------------------------------------------------------------
  -- MM buses
  ----------------------------------------------------------------------------                                         
  -- TB
  SIGNAL reg_diag_bg_mosi          : t_mem_mosi;
  SIGNAL reg_diag_bg_miso          : t_mem_miso;
                                
  SIGNAL ram_diag_bg_mosi          : t_mem_mosi;
  SIGNAL ram_diag_bg_miso          : t_mem_miso;
                                
  SIGNAL ram_diag_data_buf_re_mosi : t_mem_mosi;
  SIGNAL ram_diag_data_buf_re_miso : t_mem_miso;
  
  SIGNAL reg_diag_data_buf_re_mosi : t_mem_mosi;
  SIGNAL reg_diag_data_buf_re_miso : t_mem_miso;

  SIGNAL ram_diag_data_buf_im_mosi : t_mem_mosi;
  SIGNAL ram_diag_data_buf_im_miso : t_mem_miso;

  SIGNAL reg_diag_data_buf_im_mosi : t_mem_mosi;
  SIGNAL reg_diag_data_buf_im_miso : t_mem_miso;
  
  -- DUT              
  SIGNAL ram_ss_ss_transp_mosi     : t_mem_mosi := c_mem_mosi_rst; 
  SIGNAL ram_ss_ss_transp_miso     : t_mem_miso := c_mem_miso_rst;                           

  ----------------------------------------------------------------------------
  -- Component declaration of mm_file
  ----------------------------------------------------------------------------
  COMPONENT mm_file
  GENERIC(
    g_file_prefix       : STRING;    
    g_update_on_change  : BOOLEAN := FALSE
  );
  PORT (
    mm_rst        : IN  STD_LOGIC;
    mm_clk        : IN  STD_LOGIC;
    mm_master_out : OUT t_mem_mosi;
    mm_master_in  : IN  t_mem_miso 
  );
  END COMPONENT;   
  
  -- Compose the Constants for the DUT   
  CONSTANT c_reorder_seq_conf       : t_reorder_seq := (g_wr_chunksize, 
                                                        g_rd_chunksize, 
                                                        g_rd_nof_chunks,  
                                                        g_rd_interval,
                                                        g_gapsize,      
                                                        g_nof_blocks);   
  
  CONSTANT c_tech_ddr               : t_c_tech_ddr := c_tech_ddr3_4g_800m_master;

  CONSTANT c_min_fifo_size          : POSITIVE := 256;
  CONSTANT c_blocksize              : POSITIVE := g_wr_chunksize + g_gapsize;
  CONSTANT c_pagesize               : POSITIVE := g_nof_blocks * c_blocksize;  
  CONSTANT c_mem_size               : POSITIVE := 2*c_pagesize;
  CONSTANT c_mem_size_w             : POSITIVE := ceil_log2(c_mem_size);
                                    
  CONSTANT c_wr_fifo_depth          : NATURAL  := sel_a_b(c_blocksize > c_min_fifo_size, c_blocksize, c_min_fifo_size);--c_blocksize * 2;
  CONSTANT c_rd_fifo_depth          : NATURAL  := sel_a_b(c_blocksize > c_min_fifo_size, c_blocksize, c_min_fifo_size);--c_blocksize * 2;
                                    
                                    
  CONSTANT c_use_complex            : BOOLEAN  := TRUE;    
  CONSTANT c_total_data_w           : NATURAL  := g_nof_streams*g_in_dat_w;
  CONSTANT c_complex_data_w         : NATURAL  := c_total_data_w*c_nof_complex;
  CONSTANT c_data_w                 : NATURAL  := sel_a_b(c_use_complex, c_complex_data_w, c_total_data_w);


  CONSTANT c_ctrl_ref_clk_period    : TIME  := 5000 ps; 

  -- Custom definitions of constants
  CONSTANT c_bg_block_len           : NATURAL  := c_blocksize * g_nof_blocks;
  CONSTANT c_db_block_len           : NATURAL  := c_blocksize * g_nof_blocks;
 
  -- Configuration of the block generator:
  CONSTANT c_bg_nof_output_streams  : POSITIVE := g_nof_streams;    
  CONSTANT c_bg_buf_dat_w           : POSITIVE := c_nof_complex*g_in_dat_w;
  CONSTANT c_bg_buf_adr_w           : POSITIVE := ceil_log2(c_bg_block_len);
  CONSTANT c_bg_data_file_prefix    : STRING   := "UNUSED"; -- "../../../src/hex/tb_bg_dat";
  CONSTANT c_bg_data_file_index_arr : t_nat_natural_arr := array_init(0, 128, 1);
  
  -- Configuration of the databuffers:
  CONSTANT c_db_nof_streams         : POSITIVE := g_nof_streams;
  CONSTANT c_db_data_w              : POSITIVE := c_diag_db_max_data_w;
  CONSTANT c_db_buf_nof_data        : POSITIVE := c_db_block_len;
  CONSTANT c_db_buf_use_sync        : BOOLEAN  := FALSE;
  CONSTANT c_db_data_type_re        : t_diag_data_type_enum := e_real;
  CONSTANT c_db_data_type_im        : t_diag_data_type_enum := e_imag;

  SIGNAL bg_siso_arr                : t_dp_siso_arr(c_bg_nof_output_streams -1 DOWNTO 0) := (OTHERS => c_dp_siso_rdy);
  SIGNAL bg_sosi_arr                : t_dp_sosi_arr(c_bg_nof_output_streams -1 DOWNTO 0);

  SIGNAL out_sosi_arr               : t_dp_sosi_arr(c_bg_nof_output_streams-1 DOWNTO 0) := (OTHERS => c_dp_sosi_rst);
  SIGNAL out_siso_arr               : t_dp_siso_arr(c_bg_nof_output_streams-1 DOWNTO 0) := (OTHERS => c_dp_siso_rdy);
  
  -- Signals to interface with the DDR conroller and memory model.
  SIGNAL ctlr_dvr_miso              : t_mem_ctlr_miso;   
  SIGNAL ctlr_dvr_mosi              : t_mem_ctlr_mosi;  
  
  SIGNAL to_mem_siso                : t_dp_siso := c_dp_siso_rdy;        
  SIGNAL to_mem_sosi                : t_dp_sosi;
  SIGNAL from_mem_siso              : t_dp_siso := c_dp_siso_rdy;
  SIGNAL from_mem_sosi              : t_dp_sosi;

  SIGNAL phy_in                     : t_tech_ddr_phy_in;
  SIGNAL phy_io                     : t_tech_ddr_phy_io;
  SIGNAL phy_ou                     : t_tech_ddr_phy_ou;  
   

BEGIN

  ----------------------------------------------------------------------------
  -- Clock and reset generation
  ----------------------------------------------------------------------------
  mm_clk <= NOT mm_clk AFTER c_mm_clk_period/2;
  mm_rst <= '1', '0' AFTER c_mm_clk_period*5;

  dp_clk_in <= NOT dp_clk_in AFTER c_dp_clk_period/2;
  dp_rst_in <= '1', '0' AFTER c_dp_clk_period*5;  
  
  ------------------------------------------------------------------------------
  -- External PPS
  ------------------------------------------------------------------------------  
  proc_common_gen_pulse(1, c_dp_pps_period, '1', dp_clk, dp_pps);

   ----------------------------------------------------------------------------
  -- Procedure that polls a sim control file that can be used to e.g. get
  -- the simulation time in ns
  ----------------------------------------------------------------------------
  mmf_poll_sim_ctrl_file(c_mmf_unb_file_path & "sim.ctrl", c_mmf_unb_file_path & "sim.stat");
 
  ----------------------------------------------------------------------------
  -- MM buses  
  ----------------------------------------------------------------------------
  -- TB
  u_mm_file_reg_diag_bg          : mm_file GENERIC MAP(mmf_unb_file_prefix(0, 0, "FN") & "REG_DIAG_BG")
                                           PORT MAP(mm_rst, mm_clk, reg_diag_bg_mosi, reg_diag_bg_miso);

  u_mm_file_ram_diag_bg          : mm_file GENERIC MAP(mmf_unb_file_prefix(0, 0, "FN") & "RAM_DIAG_BG")
                                           PORT MAP(mm_rst, mm_clk, ram_diag_bg_mosi, ram_diag_bg_miso);
  
  u_mm_file_ram_diag_data_buf_re : mm_file GENERIC MAP(mmf_unb_file_prefix(0, 0, "FN") & "RAM_DIAG_DATA_BUFFER_REAL")
                                           PORT MAP(mm_rst, mm_clk, ram_diag_data_buf_re_mosi, ram_diag_data_buf_re_miso);

  u_mm_file_reg_diag_data_buf_re : mm_file GENERIC MAP(mmf_unb_file_prefix(0, 0, "FN") & "REG_DIAG_DATA_BUFFER_REAL")
                                           PORT MAP(mm_rst, mm_clk, reg_diag_data_buf_re_mosi, reg_diag_data_buf_re_miso);

  u_mm_file_ram_diag_data_buf_im : mm_file GENERIC MAP(mmf_unb_file_prefix(0, 0, "FN") & "RAM_DIAG_DATA_BUFFER_IMAG")
                                           PORT MAP(mm_rst, mm_clk, ram_diag_data_buf_im_mosi, ram_diag_data_buf_im_miso);

  u_mm_file_reg_diag_data_buf_im : mm_file GENERIC MAP(mmf_unb_file_prefix(0, 0, "FN") & "REG_DIAG_DATA_BUFFER_IMAG")
                                           PORT MAP(mm_rst, mm_clk, reg_diag_data_buf_im_mosi, reg_diag_data_buf_im_miso);

  u_mm_file_ram_ss_ss_transp     : mm_file GENERIC MAP(mmf_unb_file_prefix(0, 0, "FN") & "RAM_SS_SS_WIDE")
                                           PORT MAP(mm_rst, mm_clk, ram_ss_ss_transp_mosi, ram_ss_ss_transp_miso);

  ----------------------------------------------------------------------------
  -- Source: block generator
  ---------------------------------------------------------------------------- 
  u_bg : ENTITY diag_lib.mms_diag_block_gen
  GENERIC MAP(
    g_nof_output_streams => c_bg_nof_output_streams,
    g_buf_dat_w          => c_bg_buf_dat_w,
    g_buf_addr_w         => c_bg_buf_adr_w,            
    g_file_index_arr     => c_bg_data_file_index_arr,
    g_file_name_prefix   => c_bg_data_file_prefix
  )
  PORT MAP(
    -- System
    mm_rst               => mm_rst,
    mm_clk               => mm_clk,
    dp_rst               => dp_rst,
    dp_clk               => dp_clk,
    en_sync              => dp_pps,
    -- MM interface      
    reg_bg_ctrl_mosi     => reg_diag_bg_mosi,
    reg_bg_ctrl_miso     => reg_diag_bg_miso,
    ram_bg_data_mosi     => ram_diag_bg_mosi,
    ram_bg_data_miso     => ram_diag_bg_miso,
    -- ST interface      
    out_siso_arr         => bg_siso_arr,
    out_sosi_arr         => bg_sosi_arr
  );

  ----------------------------------------------------------------------------
  -- DUT: Device Under Test
  ---------------------------------------------------------------------------- 
  u_dut: ENTITY work.reorder_transpose
  GENERIC MAP(
    g_sim              => c_sim,   
    g_tech_ddr         => c_tech_ddr,
    g_nof_streams      => c_bg_nof_output_streams,      
    g_in_dat_w         => c_bg_buf_dat_w/c_nof_complex, 
    g_frame_size_in    => g_wr_chunksize,               
    g_frame_size_out   => g_wr_chunksize,
    g_use_complex      => c_use_complex, 
    g_mem_dat_w        => g_mem_dat_w,
    g_ena_pre_transp   => g_ena_pre_transp,                    
    g_reorder_seq      => c_reorder_seq_conf
  )                          
  PORT MAP (        
    mm_rst                => mm_rst, 
    mm_clk                => mm_clk,

    dp_rst                => dp_rst, 
    dp_clk                => dp_clk,
    
    -- ST sink                      
    snk_out_arr           => bg_siso_arr,
    snk_in_arr            => bg_sosi_arr,
    
    -- ST source          
    src_in_arr            => out_siso_arr,
    src_out_arr           => out_sosi_arr,
    
    ram_ss_ss_transp_mosi => ram_ss_ss_transp_mosi,
    ram_ss_ss_transp_miso => ram_ss_ss_transp_miso,

    -- Control interface to the external memory
    dvr_miso              => ctlr_dvr_miso,
    dvr_mosi              => ctlr_dvr_mosi,
    
    -- Data interface to the external memory
    to_mem_src_out        => to_mem_sosi,
    to_mem_src_in         => to_mem_siso,
    
    from_mem_snk_in       => from_mem_sosi,
    from_mem_snk_out      => from_mem_siso
    
  );
  
  u_ddr_mem_ctrl : ENTITY io_ddr_lib.io_ddr
  GENERIC MAP( 
    g_sim                    => c_sim,
    g_technology             => c_tech_select_default,    --  : NATURAL := c_tech_select_default;
    g_tech_ddr               => c_tech_ddr,      --  : t_c_tech_ddr;
    g_cross_domain_dvr_ctlr  => FALSE, --TRUE,   --  : BOOLEAN := TRUE;
    g_wr_data_w              => c_data_w,        --  : NATURAL := 32;  
    g_wr_fifo_depth          => c_wr_fifo_depth, --  : NATURAL := 128;     -- >=16                             , defined at DDR side of the FIFO.
    g_rd_fifo_depth          => c_rd_fifo_depth, --  : NATURAL := 256;     -- >=16 AND >g_tech_ddr.maxburstsize, defined at DDR side of the FIFO. 
    g_rd_data_w              => c_data_w,        --  : NATURAL := 32;
    g_wr_flush_mode          => "SYN",           --  : STRING := "VAL";  -- "VAL", "SOP", "SYN"
    g_wr_flush_use_channel   => FALSE,           --  : BOOLEAN := FALSE;
    g_wr_flush_start_channel => 0,               --  : NATURAL := 0;
    g_wr_flush_nof_channels  => 1                --  : POSITIVE := 1
  )                          
  PORT MAP ( 
    -- DDR reference clock
    ctlr_ref_clk  => dp_clk_in,
    ctlr_ref_rst  => dp_rst_in,

    -- DDR controller clock domain
    ctlr_clk_out  => dp_clk,   -- output clock of the ddr controller is used as DP clk.
    ctlr_rst_out  => dp_rst,   
    
    ctlr_clk_in   => dp_clk,  
    ctlr_rst_in   => dp_rst,  

    -- Driver clock domain
    dvr_clk       => dp_clk,
    dvr_rst       => dp_rst,
    
    dvr_miso      => ctlr_dvr_miso,
    dvr_mosi      => ctlr_dvr_mosi,

    -- Write FIFO clock domain
    wr_clk        => dp_clk,
    wr_rst        => dp_rst,

    wr_fifo_usedw => OPEN,
    wr_sosi       => to_mem_sosi,  
    wr_siso       => to_mem_siso,
  
    -- Read FIFO clock domain
    rd_clk        => dp_clk,
    rd_rst        => dp_rst,     
    
    rd_fifo_usedw => OPEN,
    rd_sosi       => from_mem_sosi,
    rd_siso       => from_mem_siso,     

    phy_in        => phy_in,
    phy_io        => phy_io,  
    phy_ou        => phy_ou
  );
  
  external_ddr_memory_model : IF c_sim=FALSE GENERATE
    u_tech_ddr_memory_model : ENTITY tech_ddr_lib.tech_ddr_memory_model
    GENERIC MAP (
      g_tech_ddr => c_tech_ddr
    )
    PORT MAP (
      mem_in => phy_ou,
      mem_io => phy_io
    );
  END GENERATE;

  ----------------------------------------------------------------------------
  -- Sink: data buffer real 
  ---------------------------------------------------------------------------- 
  u_data_buf_re : ENTITY diag_lib.mms_diag_data_buffer
  GENERIC MAP (    
    g_nof_streams     => c_db_nof_streams, 
    g_data_type       => c_db_data_type_re,   
    g_data_w          => c_db_data_w,      
    g_buf_nof_data    => c_db_buf_nof_data,
    g_buf_use_sync    => c_db_buf_use_sync
  )
  PORT MAP (
    -- System
    mm_rst            => mm_rst,
    mm_clk            => mm_clk,
    dp_rst            => dp_rst,
    dp_clk            => dp_clk,
     -- MM interface
    ram_data_buf_mosi => ram_diag_data_buf_re_mosi,
    ram_data_buf_miso => ram_diag_data_buf_re_miso,
    reg_data_buf_mosi => reg_diag_data_buf_re_mosi,
    reg_data_buf_miso => reg_diag_data_buf_re_miso,
    -- ST interface
    in_sync           => OPEN,
    in_sosi_arr       => out_sosi_arr
  );

  ----------------------------------------------------------------------------
  -- Sink: data buffer imag 
  ---------------------------------------------------------------------------- 
  u_data_buf_im : ENTITY diag_lib.mms_diag_data_buffer
  GENERIC MAP (    
    g_nof_streams     => c_db_nof_streams, 
    g_data_type       => c_db_data_type_im,   
    g_data_w          => c_db_data_w,      
    g_buf_nof_data    => c_db_buf_nof_data,
    g_buf_use_sync    => c_db_buf_use_sync
  )
  PORT MAP (
    -- System
    mm_rst            => mm_rst,
    mm_clk            => mm_clk,
    dp_rst            => dp_rst,
    dp_clk            => dp_clk,
    -- MM interface
    ram_data_buf_mosi => ram_diag_data_buf_im_mosi,
    ram_data_buf_miso => ram_diag_data_buf_im_miso,
    reg_data_buf_mosi => reg_diag_data_buf_im_mosi,
    reg_data_buf_miso => reg_diag_data_buf_im_miso,
    -- ST interface
    in_sync           => OPEN,
    in_sosi_arr       => out_sosi_arr
  );

END tb;
